
from __future__ import annotations

from os import path
import os
import shutil

from . import string_utils

import abc

import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

# see https://medium.com/analytics-vidhya/monitoring-your-file-system-using-watchdog-64f7ad3279f

class MonitorFolder(FileSystemEventHandler):
    def __init__(self, generator : WebsiteGenerator, folder : str, outDir : str):
        self.folder = folder
        self.outDir = outDir
        self.generator = generator
        self.observer = Observer()
        self.observer.schedule(self, self.folder, recursive=True)
        self.observer.start()

    def on_any_event(self, event):
        if event.is_directory:
            return None
        else:
            #lcdoc.run(self.folder)
            self.generator.doc_dir(self.folder, self.outDir)

    def stop(self):
        self.observer.stop()
        self.observer.join()

class WebsiteGenerator :
    """
    WebsiteGenerator
    """

    m_currentOutFolder : str = ""
    m_currentFileFolder : str = ""

    resources_folder : str = ""

    def __init__(self) -> None:
        self.resources_folder = path.join(path.dirname(__file__), "../../resources")
        pass

    def relative_href(self, href : str) -> str :
        if not self.m_currentFileFolder :
            return href
        return path.relpath(path.join(self.m_currentOutFolder, href), self.m_currentFileFolder)

    def write_rc_file(self, file : str, targetFolder : str) -> None :
        if not path.exists(targetFolder) :
            os.makedirs(targetFolder)
        shutil.copy(path.join(self.resources_folder, file), path.join(targetFolder, file))

    def html_header(self) -> str :
        return """
        <header>
            <div>
                <div class="button"><a href="">My Doc</a></div>
                <div class="links">
                    <div class="button"><a href="">References</a></div>
                    <div class="button"><a href="">Guides</a></div>
                    <div class="button"><a href="">pages</a></div>
                    <div class="button"><a href="">link</a></div>
                </div>
                <div class="button"><a href="">search</a></div>
            </div>
        </header>
        """

    def html_topnav(self) -> str :
        return """
        <lc-topnav>
            <div>
                <a href="">ciao</a>
                <a href="">ciao</a>
                <a href="">ciao</a>
            </div>
        </lc-topnav>
        """

    def html_page_base(self, title : str, content : str, additional_head_content : str = "", additional_body_content : str = "") -> str :
        return f"""
        <!DOCTYPE html>
        <html lang="en">
        <head class="light-mode">
            <meta charset="UTF-8">
            <meta http-equiv="X-UA-Compatible" content="IE=edge">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>{title}</title>

            {additional_head_content}
        </head>
        <body light-mode->
            {content}
            {additional_body_content}
        </body>
        </html>
        """

        return f"""
        <!DOCTYPE html>
        <html lang="en">
        <head class="light-mode">
            <meta charset="UTF-8">
            <meta http-equiv="X-UA-Compatible" content="IE=edge">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>Document</title>

            <!--link rel="stylesheet" href="https://lucaciucci99.com/css/style.css"-->
            <!--link rel="stylesheet" href="./style.css"-->
            <link rel="stylesheet" href="{self.relative_href("style.css")}">

            <script async src="https://cdn.jsdelivr.net/npm/tex-math@latest/dist/tex-math.js"></script>

            <script src="//cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/highlight.min.js"></script>
            <script>
                //hljs.highlightAll();
                document.addEventListener("DOMContentLoaded", function() {{
                    document.querySelectorAll('code:not(.nohighlight)').forEach((block) => {{
                        hljs.highlightElement(block);
                    }});
                }})
            </script>
            
            <script src="https://cdn.jsdelivr.net/npm/style-scoped@latest/scoped.min.js"></script>
        </head>
        <body light-mode->

            {self.html_header()}
            {self.html_topnav()}

            <lc-content>
                <article>
                    {content}
                </article>
                <lc-nav-index></lc-nav-index>
            </lc-content>

            <script src="{self.relative_href("index.js")}"></script>
        </body>
        </html>
        """
        return f"""
        <!DOCTYPE html>
        <html lang="en">
        <head class="light-mode">
            <meta charset="UTF-8">
            <meta http-equiv="X-UA-Compatible" content="IE=edge">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>Document</title>

            <link rel="stylesheet" href="https://lucaciucci99.com/css/style.css">
            <!--link rel="stylesheet" href="./style.css"-->
            <link rel="stylesheet" href="{self.relative_href("style.css")}">

            <script async src="https://cdn.jsdelivr.net/npm/tex-math@latest/dist/tex-math.js"></script>
        </head>
        <body light-mode->

            <lc-content>
                <section>
                    {content}
                </section>
            </lc-content>
        </body>
        </html>
        """

    def lcdoc_base_html(self, title : str, article : str) -> str :
        additional_head_content : str = f"""
        <!--link rel="stylesheet" href="https://lucaciucci99.com/css/style.css"-->
        <!--link rel="stylesheet" href="./style.css"-->
        <link rel="stylesheet" href="{self.relative_href("style.css")}">

        <script async src="https://cdn.jsdelivr.net/npm/tex-math@latest/dist/tex-math.js"></script>

        <script src="//cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/highlight.min.js"></script>
        <script>
            //hljs.highlightAll();
            document.addEventListener("DOMContentLoaded", function() {{
                document.querySelectorAll('code:not(.nohighlight)').forEach((block) => {{
                    hljs.highlightElement(block);
                }});
            }})
        </script>
        
        <script src="https://cdn.jsdelivr.net/npm/style-scoped@latest/scoped.min.js"></script>
        """
        body_content : str = f"""
        {self.html_header()}
        {self.html_topnav()}

        <lc-content>
            <article>
                {article}
            </article>
            <lc-nav-index></lc-nav-index>
        </lc-content>

        <script src="{self.relative_href("index.js")}"></script>
        """
        return self.html_page_base(title, body_content, additional_head_content=additional_head_content)

    def doc_dir(self, srcDir : str, outDir : str, root : bool = True) -> None :

        if root :
            self.m_currentOutFolder = outDir
            self.write_rc_file("style.css", outDir)
            self.write_rc_file("code.css", outDir)
            self.write_rc_file("index.js", outDir)

        for filename in os.listdir(srcDir) :
            in_path = path.join(srcDir, filename)
            out_path = path.join(outDir, filename)
            if path.isdir(in_path) :
                print(f"converting folder {in_path} -> {out_path}")
                self.doc_dir(in_path, out_path, False)
                continue

            if path.isfile(in_path) :
                print(f"converting file {in_path} -> {out_path}")
                content : str = ""
                with open(in_path, "r") as f :
                    content = f.read()
                    content = string_utils.recognizeCodeBlocks(content)
                p1 = content.find("---")
                p2 = content.find("---", p1 + 3)
                if p1 >= 0 and p2 >= 0 :
                    content = content[p2+3:]
                if not path.exists(path.dirname(out_path)) :
                    os.makedirs(path.dirname(out_path))
                with open(out_path, "wb") as f :
                    f.write(self.lcdoc_base_html("titolo", content).encode("utf-8"))


    def minitor_folder(self, srcDir : str, outDir : str) :
        monitor = MonitorFolder(self, srcDir, outDir)
        self.doc_dir(srcDir, outDir)
        print("running monitor")
        try:
            while True:
                time.sleep(10)
        except KeyboardInterrupt:
            monitor.stop()
            print("monitor stopped")




from dataclasses import replace
from unittest import result


def trimCommentLines(src : str) -> str :
    content : str = src.replace("/*!", "").replace("/**", "").replace("/*", "").replace("*/", "").replace("///", "").replace("//", "")

    # https://stackoverflow.com/questions/3054604/iterate-over-the-lines-of-a-string
    result : str = ""
    for line in content.splitlines() :
        strippedLine = line.strip(" ")
        if strippedLine.startswith("* ") :
            strippedLine = strippedLine[2:]
        else :
            if strippedLine.startswith("*") :
                strippedLine = strippedLine[1:]
        result += strippedLine + "\n"

    return result

def subsSpecialChars(src : str) -> str :
    result : str = src[:]
    #result = src.replace("<", "&lt;")
    #result = result.replace(">", "&gt;")
    return result

def recognizeNewLines(src : str) -> str :
    #return src.replace("  ", "<br>")
    #return src.replace("&lt;br&gt;", "<br>") # Terribile non funziona sempre
    return src.replace("\n\n", "<br>")
    return src[:]

def recognizeCodeBlocks(src : str) -> str :
    result : str = src[:]

    # ignore highlighing for code blocks withoud type
    result = result.replace("<code>", """<code class="nohighlight">""")

    while True:
        pos_1 : int = result.find("```")
        pos_2 : int = result[:].find("```", pos_1+3)
        if pos_1 >= 0 and pos_2 >= 0 :
            code = result[pos_1+3:pos_2].replace("<", "&lt;").replace(">", "&gt;")
            
            language : str = ""
            def check_language(extensions : list[str], hljs_extension : str) :
                nonlocal language, code
                for ext in extensions :
                    if code.lower().startswith(ext.lower()) :
                        language = hljs_extension
                        code = code[len(ext):]
                        break
            if not language : check_language(["cpp", "c++"], "cpp")
            if not language : check_language(["c"], "c")
            if not language : check_language(["python", "py"], "py")
            if not language : check_language(["json"], "json")
            if not language : check_language(["js", "javascript"], "js")
            if not language : check_language(["html"], "html")
            if not language : check_language(["tex", "latex"], "tex")

            code = code.strip(" \n\t")


            #result = result[:pos_1] + code + result[pos_2:]
            code_block : str = ""
            if language :
                code_block = f"""<pre><code class="language-{language}">{code}</code></pre>"""
                #result = result.replace("```", f"""<pre><code class="language-{language}">""", 1).replace("```", "</code></pre>", 1)
            else :
                code_block = f"""<pre><code class="nohighlight">{code}</code></pre>"""
                #result = result.replace("```", "<pre><code>", 1).replace("```", "</code></pre>", 1)
            result = result[:pos_1] + code_block + result[pos_2+3:]
        else :
            break

    result = result.replace("\\`", "&#96;") # wow! copilod completed with "&#96;" !!!

    return result



from __future__ import annotations

import abc
import lcdoc
from lcdoc import SymbolTable, Symbol, Declaration, TypeAliasSymbol
from typing import Iterator, Tuple
import typing
from doc_utils import DocBlock
from lcdoc import MethodSymbol
import doc_utils
from lcdoc import ClassSymbol

import string_utils

import clang.cindex

def show_list_CursorKind() -> None :
    for k in clang.cindex.CursorKind.get_all_kinds():
        print(k)

def get_cursor_kind(name : str) -> clang.cindex.CursorKind :
    return typing.cast(clang.cindex.CursorKind, getattr(clang.cindex.CursorKind, name))

CK_NAMESPACE   = get_cursor_kind("NAMESPACE")
CK_STRUCT_DECL : clang.cindex.CursorKind = clang.cindex.CursorKind.STRUCT_DECL
CK_CLASS_DECL  : clang.cindex.CursorKind = clang.cindex.CursorKind.CLASS_DECL
CK_CXX_METHOD  : clang.cindex.CursorKind = clang.cindex.CursorKind.CXX_METHOD
CK_TYPE_ALIAS_DECL  : clang.cindex.CursorKind = clang.cindex.CursorKind.TYPE_ALIAS_DECL

TK_VOID  : clang.cindex.TypeKind = clang.cindex.TypeKind.VOID
TK_INT  : clang.cindex.TypeKind = clang.cindex.TypeKind.INT
TK_RVALUEREFERENCE  : clang.cindex.TypeKind = clang.cindex.TypeKind.RVALUEREFERENCE
TK_LVALUEREFERENCE  : clang.cindex.TypeKind = clang.cindex.TypeKind.LVALUEREFERENCE
TK_POINTER  : clang.cindex.TypeKind = clang.cindex.TypeKind.POINTER
TK_ELABORATED  : clang.cindex.TypeKind = clang.cindex.TypeKind.ELABORATED
TK_AUTO  : clang.cindex.TypeKind = clang.cindex.TypeKind.AUTO

def get_access_specifier(name : str) -> clang.cindex.AccessSpecifier :
    return typing.cast(clang.cindex.AccessSpecifier, getattr(clang.cindex.AccessSpecifier, name))

AS_INVALID   = get_access_specifier("INVALID")
AS_PUBLIC    = get_access_specifier("PUBLIC")
AS_PROTECTED = get_access_specifier("PROTECTED")
AS_PRIVATE   = get_access_specifier("PRIVATE")
AS_NONE      = get_access_specifier("NONE")

class Node(lcdoc.Node) :

    cursor : clang.cindex.Cursor

    def __init__(self, cursor : clang.cindex.Cursor) -> None:
        super().__init__()

        self.cursor = cursor
    
    def spelling(self) -> str :
        return self.cursor.spelling

    def displayname(self) -> str :
        return self.cursor.displayname

    def file(self) -> str :
        return typing.cast(clang.cindex.SourceLocation, self.cursor.location).file.name
        
class Type :

    const_qualified : bool = False

    def __init__(self) -> None:
        pass

    @abc.abstractclassmethod
    def __repr__(self) -> str :
        pass

    @abc.abstractclassmethod
    def to_html(self, context : Symbol | None = None) -> str :
        pass

    def get_html_const_prefix(self) -> str :
        if self.const_qualified :
            return "<code-keyword>const</code-keyword> "
        return ""

class BasicType(Type) :
    ty : str = ""

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return self.ty

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.ty
        return self.get_html_const_prefix() + f"<code-keyword>{self.ty}</code-keyword>"

class RValueRefType(Type) :
    to : Type | None = None

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        if self.to :
            return self.to.__repr__() + "&&"
        return "ERROR &&"

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.to
        return self.get_html_const_prefix() + f"{self.to.to_html(context)}&&"

class LValueRefType(Type) :
    to : Type | None = None

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        if self.to :
            return self.to.__repr__() + "&"
        return "ERROR &"

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.to
        return self.get_html_const_prefix() + f"{self.to.to_html(context)}&"

class PointerType(Type) :
    to : Type | None = None

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        if self.to :
            return self.to.__repr__() + "*"
        return "ERROR *"

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.to
        if self.const_qualified :
            return self.to.to_html(context) + " " + self.get_html_const_prefix() + "*"
        return f"{self.to.to_html(context)}*"

class ClassType(Type) :

    classSymbol : ClassSymbol | None = None

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        if self.classSymbol :
            return self.classSymbol.__repr__()
        return "ERROR"

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.classSymbol
        return self.get_html_const_prefix() + f"{self.classSymbol.to_html_full(context)} &"

class TypeAliasType(Type) :

    aliasSymbol : lcdoc.TypeAliasSymbol | None = None

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        if self.aliasSymbol :
            return self.aliasSymbol.__repr__()
        return "ERROR"

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.aliasSymbol
        return self.get_html_const_prefix() + self.aliasSymbol.to_html_full(context)

class AutoType(Type) :

    deduced : Symbol | None = None

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        if self.deduced :
            return self.deduced.__repr__()
        return "ERROR"

    def to_html(self, context : Symbol | None = None) -> str :
        assert self.deduced
        return self.get_html_const_prefix() + self.deduced.to_html_full(context)

class ParsedDocument :

    filePath : str = ""

    symbolTable : SymbolTable

    declarations : list[Declaration]

    def __init__(self) -> None:
        self.symbolTable = SymbolTable()
        self.declarations = []

class Definition :

    defName : str = ""
    defValue : str = ""

    def __init__(self) -> None:
        pass

class FileOptions :
    includeDirs : list[str]
    definitions : list[Definition]

    def __init__(self) -> None:
        self.includeDirs = []
        self.definitions = []

    def merge(self, child : FileOptions) -> FileOptions :
        result = FileOptions()

        for i in self.includeDirs :
            result.includeDirs.append(i)
        for i in child.includeDirs :
            result.includeDirs.append(i)

        for d in self.definitions :
            result.definitions.append(d)
        for d in child.definitions :
            result.definitions.append(d)

        return result

class DocumentParser :

    includes : list[str]
    inFileOnly : bool = False

    m_filePath : str
    m_currDoc : ParsedDocument

    def __init__(self) -> None:
        self.includes = []

    def parse(self, filePath : str) -> ParsedDocument :

        self.m_filePath = filePath

        # https://clang.llvm.org/docs/ClangCommandLineReference.html
        args : list[str] = []
        for include in self.includes :
            args.append("-I%s" % include)
        args.append("--std=c++20")

        index : clang.cindex.Index = clang.cindex.Index.create()
        tu : clang.cindex.TranslationUnit = index.parse(filePath, args)
        #for d in tu.diagnostics.tu.cursor.get_children() :
        #    print(typing.cast(clang.cindex.Cursor, d).spelling, type(d))

        # The document Abstract Syntax Tree (AST)
        AST : list[clang.cindex.Cursor] = []
        for node in tu.cursor.get_children() :
            AST.append(node)

        doc = ParsedDocument()
        self.m_currDoc = doc
        doc.filePath = filePath

        self.findDeclarations(doc, AST)
        
        return doc

    def findDeclarations(self, doc : ParsedDocument, AST : list[clang.cindex.Cursor]) -> None :
        doc.declarations.clear()
        self.findDeclarations_impl(doc, AST)

    def findDeclarations_impl(self, doc : ParsedDocument, nodes : Iterator[clang.cindex.Cursor] | list[clang.cindex.Cursor], parent : Symbol | None = None) -> None :

        for node in nodes :
            #print(node.kind, node.spelling)

            if not node.spelling :
                # this because sometime "Unknown template argument kind 280" ValueError
                # exception i thrown
                continue

            if self.inFileOnly :
                location = typing.cast(clang.cindex.SourceLocation, node.location)
                assert location.file
                if location.file.name != self.m_filePath :
                    continue
            
            if node.kind == CK_NAMESPACE :
                symbol = self.find_symbol_or_add(doc, node, lcdoc.NamespaceSymbol, parent)
                self.findDeclarations_impl(doc, node.get_children(), symbol)
                continue

            if node.kind == CK_STRUCT_DECL :
                self.find_symbol_and_decl_or_add(doc, node, lcdoc.StructDeclaration, lcdoc.StructSymbol, parent)
                continue

            if node.kind == CK_CLASS_DECL :
                self.find_symbol_and_decl_or_add(doc, node, lcdoc.ClassDeclaration, lcdoc.ClassSymbol, parent)
                continue

            if node.kind == CK_CXX_METHOD :
                symbol, decl = self.find_symbol_and_decl_or_add(doc, node, lcdoc.MethodDeclaration, lcdoc.MethodSymbol, parent)
                symbol = typing.cast(lcdoc.MethodSymbol, symbol)
                symbol.ret = FuncArg()
                #symbol.ret.strType = typing.cast(clang.cindex.Type, typing.cast(clang.cindex.Type, node.result_type).get_canonical()).spelling
                symbol.ret.strType = typing.cast(clang.cindex.Type, node.result_type).spelling
                symbol.ret.typ = self.clangType_to_Type(typing.cast(clang.cindex.Type, node.result_type))
                AS = typing.cast(clang.cindex.AccessSpecifier, node.access_specifier)
                if AS == AS_PRIVATE :
                    symbol.access_specifier = lcdoc.AccessSpecifier.PRIVATE
                if AS == AS_PROTECTED :
                    symbol.access_specifier = lcdoc.AccessSpecifier.PROTECTED
                if AS == AS_PUBLIC :
                    symbol.access_specifier = lcdoc.AccessSpecifier.PUBLIC
                if AS == AS_INVALID or AS == AS_NONE :
                    assert False
                symbol.const_qualified = node.is_const_method()
                for a in node.get_arguments() :
                    arg = FuncArg()
                    argument = typing.cast(clang.cindex.Cursor, a)
                    argType = typing.cast(clang.cindex.Type, argument.type)
                    arg.name = argument.spelling
                    #arg.strType = typing.cast(clang.cindex.Type, argType.get_canonical()).
                    arg.strType = argType.spelling
                    arg.typ = self.clangType_to_Type(argType)
                    symbol = typing.cast(lcdoc.MethodSymbol, symbol)
                    symbol.args.clear()
                    symbol.args.append(arg)
                continue

            if node.kind == CK_TYPE_ALIAS_DECL :
                symbol, decl = self.find_symbol_and_decl_or_add(doc, node, lcdoc.TypeAliasDeclaration, lcdoc.TypeAliasSymbol, parent)
                # TODO TO WHAT????
                continue

            self.findDeclarations_impl(doc, node.get_children(), parent)

    def find_symbol_or_add(self, doc : ParsedDocument, node : clang.cindex.Cursor, symbol_type : type, parent : Symbol | None) -> Symbol :
        symbol = doc.symbolTable.find(node.mangled_name if node.mangled_name else node.spelling, symbol_type, parent)
        symbol = typing.cast(lcdoc.Symbol, symbol)
        if symbol == None :
            symbol = symbol_type()
            symbol = typing.cast(lcdoc.Symbol, symbol)
            symbol.mangledName = node.mangled_name
            symbol.name = node.spelling
            if not symbol.mangledName :
                symbol.mangledName = symbol.name
            if parent != None :
                symbol.addToParent(parent)
            doc.symbolTable.add(symbol)
        return symbol

    def find_symbol_and_decl_or_add(self, doc : ParsedDocument, node : clang.cindex.Cursor, decl_type : type, symbol_type : type, parent : Symbol | None) -> Tuple[ Symbol, Declaration ] :
        symbol = self.find_symbol_or_add(doc, node, symbol_type, parent)
        decl : Declaration | None = None
        for d in doc.declarations :
            if d.symbol == symbol :
                decl = d
                break
        if decl == None :
            decl = typing.cast(Declaration, decl_type(Node(node)))
            decl.symbol = symbol
            doc.declarations.append(decl)
        symbol.declaration = decl
        symbol.definition = decl
        #symbol.documentationString.brief =  ("brief: " + str(node.brief_comment)) if node.brief_comment else ""
        #symbol.documentationString.docStr += ("raw: " + string_utils.trimCommentLines(str(node.raw_comment))) if node.raw_comment else "" + "\n"
        symbol.documentationString.brief =  str(node.brief_comment) if node.brief_comment else ""
        symbol.documentationString.docStr += string_utils.trimCommentLines(str(node.raw_comment)) if node.raw_comment else "" + "\n"
        self.findDeclarations_impl(doc, node.get_children(), symbol)

        return symbol, decl

    def clangType_to_Type(self, cType : clang.cindex.Type) -> Type | None :
        if cType.kind == TK_VOID :
            result = BasicType()
            result.ty = "void"
            return result
        
        if cType.kind == TK_INT :
            result = BasicType()
            result.ty = "int"
            result.const_qualified = cType.is_const_qualified()
            return result
        
        if cType.kind == TK_RVALUEREFERENCE :
            ref = RValueRefType()
            ref.to = self.clangType_to_Type(cType.get_pointee())
            ref.const_qualified = cType.is_const_qualified()
            return ref

        if cType.kind == TK_LVALUEREFERENCE :
            ref = LValueRefType()
            ref.to = self.clangType_to_Type(cType.get_pointee())
            ref.const_qualified = cType.is_const_qualified()
            return ref

        if cType.kind == TK_POINTER :
            ptr = PointerType()
            ptr.to = self.clangType_to_Type(cType.get_pointee())
            ptr.const_qualified = cType.is_const_qualified()
            return ptr

        if cType.kind == TK_ELABORATED :
            cl = TypeAliasType()
            symbol = self.findSybolByCursor(cType.get_declaration())
            assert symbol
            cl.aliasSymbol = typing.cast(TypeAliasSymbol, symbol)
            cl.const_qualified = cType.is_const_qualified()
            return cl

        if cType.kind == TK_AUTO :
            au = AutoType()
            au.deduced = self.findSybolByCursor(cType.get_declaration())
            au.const_qualified = cType.is_const_qualified()
            return au

        print("unknown kind: ", cType.kind)
        
        return None

    def findSybolByCursor(self, cursor : clang.cindex.Cursor) -> Symbol | None :
        for symbol in self.m_currDoc.symbolTable.symbols :
            if symbol.definition :
                if cursor == typing.cast(Node, symbol.definition.node).cursor :
                    return symbol
        return None



class FuncArg(lcdoc.FuncArg) :
    strType : str = ""
    typ : Type | None = None
    pass

    def to_html_full(self) -> str :
        assert self.typ
        return self.typ.to_html() + " " + self.name



class Generator :

    symbolTable : SymbolTable

    def __init__(self) -> None:
        self.symbolTable = SymbolTable()

        # TODO REMOVE
        symbolTable = self.symbolTable
        std = lcdoc.NamespaceSymbol()
        std.name = "std"
        stringSymbol = lcdoc.ClassSymbol()
        stringSymbol.name = "basic_string"
        stringSymbol.addToParent(std)
        symbolTable.add(std)
        symbolTable.add(stringSymbol)

    def html(self, doc : ParsedDocument) :

        # https://stackoverflow.com/questions/46003452/how-to-correctly-write-a-raw-multiline-string-in-python
        # https://stackoverflow.com/questions/10660435/pythonic-way-to-create-a-long-multi-line-string
        html = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Document</title>

    <link rel="stylesheet" href="https://lucaciucci99.com/css/style.css">
    <link rel="stylesheet" href="./style.css">

    <script async src="https://cdn.jsdelivr.net/npm/tex-math@latest/dist/tex-math.js"></script>
</head>
<body>

    <lc-content>
        <section>
            <article>
        """

        for symbol in doc.symbolTable.symbols :
            if isinstance(symbol, lcdoc.ClassSymbol) :
                html += "<article-page>%s</article-page>" % self.class_to_html_page(symbol)

        if False :
            for symbol in doc.symbolTable.symbols :
                if isinstance(symbol, lcdoc.ClassSymbol) or isinstance(symbol, lcdoc.StructSymbol) :
                    html += """
    <article-page>
    %s
    </article-page>
                    """ % self.class_to_html(symbol)
                    html += "\n"
            
            html += """
                </article>
            </section>
        </lc-content>

    </body>
    </html>
            """

        return html

    def class_to_html_page(self, classSymbol : lcdoc.ClassSymbol) -> str :
        html : str = ""

        classDoc = doc_utils.DocBlock()
        classDoc.parse(classSymbol.documentationString.docStr)

        # title
        html += f"<h1>{classSymbol.name}</h1>"

        # brief description
        html += f"<h2>{classSymbol.to_html_name()}</h2><hr>"
        html += f"<p>{classDoc.brief}</p>"
        #html += f"<p>{classSymbol.getCompleteName()}</p>"
        html += f"<p>Full Symbol:<pre><code>{classSymbol.to_html_full()}</code></pre></p>"

        # methods
        html += "<h2>Methods</h2>"
        rows : str = ""
        for child in classSymbol.children :
            if isinstance(child, MethodSymbol) :
                method : MethodSymbol = typing.cast(MethodSymbol, child)
                assert method.declaration
                node = typing.cast(Node, method.declaration.node)
                cursor = node.cursor
                row : str = ""
                #row += self.html_table_data(f"{self.symbol_str_highlight_to_html(typing.cast(FuncArg, method.ret).strType)}", "return-type")
                row += self.html_table_data(f"<code>{typing.cast(FuncArg, method.ret).typ.to_html()}</code>", "return-type")
                args : str = ""
                for arg in method.args[:-1] :
                    arg2 = typing.cast(FuncArg, arg)
                    args += f"{self.type_to_html(arg2.typ)} <code-mvar>{arg2.name}</code-mvar>, "
                if len(method.args) > 0 :
                    arg = method.args[-1]
                    arg2 = typing.cast(FuncArg, arg)
                    args += f"{arg2.typ.to_html()} <code-mvar>{arg2.name}</code-mvar>"
                doc = doc_utils.DocBlock()
                doc.parse(method.documentationString.docStr)
                row += self.html_table_data(f"""
<code><code-function>{method.name}</code-function>({args}){" <code-keyword>const</code-keyword>" if cursor.is_const_method() else ""}</code><br>{self.method_doc_to_html_details(doc, method)}
                """)
                rows += self.html_table_row(row)
        html += self.html_table(rows, "basic-table")

        return html

    def method_doc_to_html_details(self, doc : doc_utils.DocBlock, methodSymbol : lcdoc.MethodSymbol) -> str :
        if not doc.long :
            return doc.brief

        return f"""
<details>
    <summary>{doc.brief}</summary>
    {doc.long}
    {self.args_html_table_section(methodSymbol.args, doc)}
</details>
            """

    def args_html_table_section(self, args : list[lcdoc.FuncArg], methodDoc : doc_utils.DocBlock) -> str :
        if len(args) <= 0 :
            return ""
        return f"""
<h4>Parameters</h4>
{self.args_to_html_table(args, methodDoc)}
        """

    def args_to_html_table(self, args : list[lcdoc.FuncArg], methodDoc : doc_utils.DocBlock) -> str :
        rows : str = ""
        for _arg in args :
            arg = typing.cast(FuncArg, _arg)
            row : str = ""
            #row += self.html_table_data(f"<code>{arg.strType}</code>", "type")
            row += self.html_table_data(f"<code>{arg.typ.to_html()}</code>", "type")
            tmp = ""
            paramDocBlock = methodDoc.findParam(arg.name)
            if paramDocBlock :
                doc = doc_utils.DocBlock()
                doc.parse(paramDocBlock.descr)
                tmp = f"<details><summary>{doc.brief}</summary>{doc.long}</details>"
            row += self.html_table_data(f"<code><code-mvar>{arg.name}</code-mvar></code>{tmp}")
            rows += self.html_table_row(row)
        return self.html_table(rows, "basic-table")

    def html_table(self, content : str, classes : str = "") -> str :
        return f"""
<table class="{classes}">
{content}
</table>
        """

    def html_table_row(self, content : str, classes : str = "") -> str :
        return f"""
<tr class="{classes}">
{content}
</tr>
        """

    def html_table_data(self, content : str, classes : str = "") -> str :
        return f"""
<td class="{classes}">
{content}
</td>
        """

    def symbol_str_highlight_to_html(self, src : str) -> str :
        html : str = ""
        pieces = src.split()
        for piece in pieces :
            # keywords
            if self.isKeyword(piece) :
                html += f"<code-keyword>{piece}</code-keyword> "
                continue

            # symbol
            ss = piece.split("::")
            tot : str = ""
            for i in range(len(ss)) :
                symbol = self.findSymbol("::".join(ss[:i+1]))
                if not symbol :
                    if tot :
                        tot += "::"
                    tot += "::".join(ss[i:])
                    break
                if tot :
                    tot += "::"
                tot += self.get_symbol_html_name(symbol)
            if tot :
                html += tot
                continue

            # everything else
            html += piece + " "
        return html

    # TODO in which context and global if starts with ::
    def findSymbol(self, src : str) -> Symbol | None :
        if src.find("::") >= 0 :
            pass
        for symbol in self.symbolTable.symbols :
            if symbol.getCompleteName() == src :
                return symbol
        return None

    def isKeyword(self, name : str) -> bool :
        if name == "true" :
            return True
        if name == "false" :
            return True
        if name == "void" :
            return True
        if name == "int" :
            return True
        if name == "unsigned" :
            return True
        if name == "long" :
            return True
        if name == "const" :
            return True
        return False




# https://stackoverflow.com/questions/19079070/retrieving-comments-using-python-libclang



from __future__ import annotations

import cxx
import lcdoc
from os import path

class CxxTargetOptions :

    fileOpt : cxx.FileOptions
    
    def __init__(self) -> None:
        self.fileOpt = cxx.FileOptions()

class CxxTargetFile :

    fileName : str = ""
    additionalOptions : CxxTargetOptions
    
    def __init__(self) -> None:
        self.additionalOptions = CxxTargetOptions()

class CxxTarget :

    options : CxxTargetOptions

    files : list[CxxTargetFile]

    def __init__(self) -> None:
        self.options = CxxTargetOptions()
        self.files = []

    def add_file(self, fileName : str) :
        file = CxxTargetFile()
        file.fileName = fileName
        self.files.append(file)





from mimetypes import common_types
import string_utils

class ParamDocBlock :
    name : str = ""
    descr : str = ""

    def handle(self, src : str) :
        comment = src.strip()
        pos : int = comment.find(" ")
        if (pos >= 0) :
            self.name = comment[:pos]
            self.descr = comment[pos:].strip()
        else :
            self.name = src

class DocBlock :

    brief : str = ""
    long : str = ""

    params : list[ParamDocBlock]

    def __init__(self) -> None:
        self.params = []
        pass

    def findParam(self, name : str) -> ParamDocBlock | None :
        for param in self.params :
            if (param.name == name) :
                return param
        return None
    
    def parse(self, src : str) -> None :
        #comment = string_utils.trimCommentLines(src)
        comment = src[:]

        # Find the brief block
        briefEnd = comment.find("\n\n")
        if briefEnd > 0 :
            self.brief = comment[:briefEnd]
            comment = comment[briefEnd+2:]
        else :
            self.brief = comment
            comment = ""
        self.brief = self.brief.replace("@brief ", "", 1)
        self.brief = self.brief.replace("\\brief ", "", 1)

        #print("brief: ", self.brief)
        #print("long: ", comment)

        while comment :
            lineEnd : int = comment.find("\n")
            line : str = comment[:]
            if lineEnd >= 0 :
                line = comment[:lineEnd]
            if line.startswith("@") :
                self.handleDifferentBlock(comment)
                break

            if lineEnd >= 0 :
                comment = comment[lineEnd+1:]
            else :
                comment = ""
            self.long += line + "\n"
        
        self.long = string_utils.recognizeCodeBlocks(self.long).replace("<p>", "<div class=\"paragraph\">").replace("</p>", "</div>")

    def handleDifferentBlock(self, src : str) -> None :
        if not src :
            return
        if src[1:].startswith("param ") :
            param = ParamDocBlock()
            self.params.append(param)
            param.handle(self.handleBlock(src[1:].replace("param ", "", 1)))
            pass

    def handleBlock(self, src : str) -> str :
        comment = src[:]
        result = ""
        while comment :
            lineEnd : int = comment.find("\n")
            line : str = comment[:]
            if lineEnd >= 0 :
                line = comment[:lineEnd]
            if line.startswith("@") :
                self.handleDifferentBlock(comment)
                break
            if lineEnd >= 0 :
                comment = comment[lineEnd+1:]
            else :
                comment = ""
            result += line + "\n"
        return result



"""LC_NOTICE_BEGIN
LC_NOTICE_END"""


from __future__ import annotations
import abc
from enum import Enum
from operator import contains

from typing import Iterator

class DocumentationString :
    brief : str = ""
    docStr : str = ""

class Symbol :
    declaration : Declaration | None = None
    definition : Declaration | None = None

    name : str = ""
    mangledName : str = ""

    documentationString : DocumentationString

    parent : Symbol | None = None
    children : list[Symbol]

    def __init__(self) -> None:
        self.children = []
        self.documentationString = DocumentationString()

    def addToParent(self, parent : Symbol) -> None :
        if not contains(parent.children, self) :
            parent.children.append(self)
        self.parent = parent

    def getCompleteName(self) -> str :
        if (self.parent) :
            return self.parent.getCompleteName() + "!!" + self.name
        return self.name

    @abc.abstractclassmethod
    def to_html_name(self) -> str :
        pass

    def to_html_full(self, context : Symbol | None = None) -> str :
        if self.parent :
            if context :
                if self.parent == context :
                    return self.to_html_name()
            return self.parent.to_html_full() + "::" + self.to_html_name()
        return self.to_html_name()

class NamespaceSymbol(Symbol) :

    def getCompleteName(self) -> str :
        if isinstance(self.parent, NamespaceSymbol) :
            return self.parent.getCompleteName() + "::" + self.name
        return self.name

    def __repr__(self) -> str:
        return "namespace " + self.getCompleteName()

    def to_html_name(self) -> str :
        return f"<code-namespace>{self.name}</code-namespace>"

# abstract
class TypeSymbol(Symbol) :
    pass

class StructSymbol(TypeSymbol) :

    def getCompleteName(self) -> str :
        if isinstance(self.parent, NamespaceSymbol) :
            return self.parent.getCompleteName() + "::" + self.name
        return self.name

    def __repr__(self) -> str:
        return "struct " + self.getCompleteName()

    def to_html_name(self) -> str :
        return f"<code-class>{self.name}</code-class>"

class ClassSymbol(TypeSymbol) :

    def getCompleteName(self) -> str :
        if isinstance(self.parent, NamespaceSymbol) :
            return self.parent.getCompleteName() + "::" + self.name
        return self.name

    def __repr__(self) -> str:
        return "class " + self.getCompleteName()

    def to_html_name(self) -> str :
        return f"<code-class>{self.name}</code-class>"

class FuncArg :
    name : str = ""

    @abc.abstractmethod
    def to_html_full(self, context : Symbol | None = None) -> str :
        return f"\"{self.name}\""

class AccessSpecifier(Enum) :
    PRIVATE   =  0
    PUBLIC    = 1
    PROTECTED = 2

class MethodSymbol(Symbol) :

    args : list[FuncArg]

    ret : FuncArg

    access_specifier : AccessSpecifier
    
    const_qualified : bool = False

    def __init__(self) -> None:
        super().__init__()
        self.args = []
        self.ret = FuncArg()
        self.access_specifier = AccessSpecifier(AccessSpecifier.PUBLIC)

    def getCompleteName(self) -> str :
        if isinstance(self.parent, NamespaceSymbol) :
            return self.parent.getCompleteName() + "::" + self.name
        return self.name

    def __repr__(self) -> str:
        return "method " + self.getCompleteName()

    def to_html_name(self) -> str :
        return f"<code-function>{self.name}</code-function>"

    def to_html_full(self, context : Symbol | None = None) -> str :
        r : str = self.ret.to_html_full()
        # TODO params
        if self.parent :
            if context :
                if self.parent == context :
                    return r + " " + self.to_html_name()
            return r + " " + self.parent.to_html_full() + "::" + self.to_html_name()
        return r + " " + self.to_html_name()

class TypeAliasSymbol(TypeSymbol) :
    to : TypeSymbol | None = None

    def getCompleteName(self) -> str :
        if self.parent :
            return self.parent.getCompleteName() + "::" + self.name
        return self.name

    def __repr__(self) -> str:
        return "alias " + self.getCompleteName() + " to " + (self.to.__repr__() if self.to else "ERROR")

    def to_html_name(self) -> str :
        return f"<code-class>{self.name}</code-class>"

class SymbolTable :
    symbols : list[Symbol]
    root_symbols : list[Symbol]

    def __init__(self) -> None:
        self.symbols = []
        self.root_symbols = []

    def add(self, symbol : Symbol) -> None :
        self.symbols.append(symbol)
        if (symbol.parent == None) :
            self.root_symbols.append(symbol)

    def find(self, mangledName : str, class_type : type, parent : Symbol | None = None) -> Symbol | None :
        for symbol in self.symbols :
            if mangledName == symbol.mangledName and isinstance(symbol, class_type) :
                if parent :
                    if symbol.parent == parent :
                        return symbol
                    else :
                        return None
                return symbol
        return None

class Node :

    @abc.abstractclassmethod
    def spelling(self) -> str :
        pass

    @abc.abstractclassmethod
    def displayname(self) -> str :
        pass

    @abc.abstractclassmethod
    def file(self) -> str :
        pass


class Declaration :
    ciao : str = "aa"
    node : Node

    symbol : Symbol | None = None

    def __init__(self, node : Node) -> None :
        self.node = node

    def spelling(self) -> str :
        return self.node.spelling()

    def __str__(self) -> str:
        return self.__repr__()
        #return str(self.node.kind) + " " + self.node.spelling()

    def __repr__(self) -> str:
        # see https://docs.python.org/3/reference/datamodel.html#object.__repr__
        #return self.node.displayname()
        return self.node.spelling()

class StructDeclaration(Declaration) :

    def __init__(self, node : Node) -> None :
        super().__init__(node)

class ClassDeclaration(Declaration) :

    def __init__(self, node : Node) -> None :
        super().__init__(node)

class MethodDeclaration(Declaration) :

    def __init__(self, node : Node) -> None :
        super().__init__(node)

class TypeAliasDeclaration(Declaration) :

    def __init__(self, node : Node) -> None:
        super().__init__(node)




from __future__ import annotations
from os import link, path
import os

from cxx import *
from lcdoc import *
import cxx
from doc_utils import ParamDocBlock
from WebsiteGenerator import WebsiteGenerator

TAG_NAMESPACE = "code-namespace"
TAG_CLASS = "code-class"
TAG_FUNCTION = "code-function"

class SymbolsTree :

    symbol : Symbol | None = None
    children : list[SymbolsTree]

    def __init__(self) -> None:
        self.children = []

    def insertSymbolTree(self, root : Symbol, fileName : str) -> SymbolsTree :
        subTree : SymbolsTree | None = None
        
        # find the correct subtree
        for child in self.children :
            if root.mangledName == typing.cast(Symbol, child.symbol).mangledName :
                subTree = child
                break
        
        # if not found, add a subtree
        if not subTree :
            subTree = SymbolsTree()
            subTree.symbol = root
            self.children.append(subTree)

        # if it is declared in this file, force substitution
        if root.declaration :
            # TODO merge docs
            if root.declaration.node.file() == fileName :
                subTree.symbol = root

        for childSymbol in root.children :
            #print(childSymbol)
            subTree.insertSymbolTree(childSymbol, fileName)

        return subTree

    def fixParents(self) -> None :
        for childTree in self.children :
            if self.symbol :
                assert childTree.symbol
                childTree.symbol.addToParent(self.symbol)
            childTree.fixParents()


class HtmlGenerator(WebsiteGenerator) :

    symbolTree : SymbolsTree

    def __init__(self) -> None:
        super().__init__()
        self.symbolTree = SymbolsTree()

    def add_doc(self, doc : ParsedDocument) -> None :
        for root in doc.symbolTable.root_symbols :
            self.symbolTree.insertSymbolTree(root, doc.filePath)
            self.symbolTree.fixParents()
        pass

    def generate_symbol_href(self, symbol : Symbol, final : bool = True) -> str :
        href : str = ""

        pre_v : str = "symbols"
        if symbol.parent :
            pre_v = self.generate_symbol_href(symbol.parent, False)

        if isinstance(symbol, NamespaceSymbol) :
            if final :
                return self.relative_href(path.join(pre_v, symbol.name))
            return path.join(pre_v, symbol.name)

        if isinstance(symbol, StructSymbol) :
            if final :
                return self.relative_href(path.join(pre_v, symbol.name))
            return path.join(pre_v, symbol.name)

        if isinstance(symbol, ClassSymbol) :
            if final :
                return self.relative_href(path.join(pre_v, symbol.name))
            return path.join(pre_v, symbol.name)

        if isinstance(symbol, TypeAliasSymbol) :
            if final :
                return self.relative_href(path.join(pre_v, symbol.name))
            return path.join(pre_v, symbol.name)

        if isinstance(symbol, MethodSymbol) :
            if final :
                return path.join(self.relative_href(pre_v), "#" + self.generete_symbol_id(symbol).replace("?", "%3f"))
            return "#".join([pre_v, symbol.name])

        return "#ERROR"

    def generete_symbol_id(self, symbol : Symbol, pre : str = "") -> str :
        pre_v : str = pre[:]
        #if symbol.parent :
        #    pre_v += self.generete_symbol_id() + "::"

        return pre_v + symbol.mangledName

    def generate_keyword_href(self, ky : str) -> str| list[str] :
        # see https://en.cppreference.com/w/cpp/keyword
        # https://en.cppreference.com/w/cpp/language/structured_binding

        if ky == "const" :
            return [
                "https://en.cppreference.com/w/cpp/keyword/const",
                "https://en.cppreference.com/book/intro/const"
            ]

        if ky == "int" :
            return [
                "https://en.cppreference.com/w/cpp/keyword/int",
                "https://en.cppreference.com/w/cpp/language/types",
            ]

        if ky == "void" :
            return [
                "https://en.cppreference.com/w/cpp/keyword/void",
                "https://en.cppreference.com/w/cpp/language/types",
            ]

        if ky == "class" :
            return [
                "https://en.cppreference.com/w/cpp/keyword/class",
                "https://en.cppreference.com/w/cpp/language/class",
                #"https://en.cppreference.com/w/cpp/language/classes",
                #"https://en.cppreference.com/book/intro/classes",
                #"https://en.cppreference.com/book/class"
            ]

        if ky == "auto" :
            return [
                "https://en.cppreference.com/w/cpp/keyword/auto",
                "https://en.cppreference.com/w/cpp/language/auto",
                "https://docs.microsoft.com/en-us/cpp/cpp/auto-cpp?view=msvc-170"
            ]


        if ky == "&&" or ky == "&" :
            return [
                "https://en.cppreference.com/w/cpp/language/reference",
                "https://en.cppreference.com/book/intro/reference"
            ]

        if ky == "*" :
            return [
                "https://en.cppreference.com/w/cpp/language/pointer",
                "https://en.cppreference.com/book/pointers"
            ]

        return ""

    def tag(self, tagName : str, content : str, data : str = "") -> str :
            if data :
                return f"<{tagName} {data}>{content}</{tagName}>"
            return f"<{tagName}>{content}</{tagName}>"

    def code_tag(self, content : str) -> str :
        return self.tag("code", content, 'class="nohighlight"')
    
    def code_span_tag(self, tagName : str, content : str, href : str | list[str] = "") -> str :
        if isinstance(href, list) :
            links : str = ""
            for h in href :
                links += self.tag("li", self.tag("a", h, f'href="{h}"'))
            tooltiptext : str = self.tag("span", "links:" + self.tag("ol", links), 'class="tooltiptext"')
            # TODO change
            return self.tag(tagName, content + tooltiptext, 'class="tooltip"')

        if href :
            return f"<a class=\"code-link\" href=\"{href}\"><{tagName}>{content}</{tagName}></a>"
        return f"<{tagName}>{content}</{tagName}>"

    def symbolHtmlName(self, symbol : Symbol) :

        if isinstance(symbol, NamespaceSymbol) :
            return self.code_span_tag(TAG_NAMESPACE, symbol.name, self.generate_symbol_href(symbol))

        if isinstance(symbol, StructSymbol) or isinstance(symbol, ClassSymbol) :
            return self.code_span_tag(TAG_CLASS, symbol.name, self.generate_symbol_href(symbol))
        
        if isinstance(symbol, MethodSymbol) :
            return self.code_span_tag(TAG_FUNCTION, symbol.name, self.generate_symbol_href(symbol))

        if isinstance(symbol, TypeAliasSymbol) :
            return self.code_span_tag(TAG_CLASS, symbol.name, self.generate_symbol_href(symbol))

        return "symbolHtmlName_ERROR"

    def symbolCompleteHtmlName(self, symbol : Symbol) -> str :
        if symbol.parent :
            return self.symbolCompleteHtmlName(symbol.parent) + "::" + self.symbolHtmlName(symbol)
        return self.symbolHtmlName(symbol)

    # TODO conetxt
    def type2Html(self, ty : Type) -> str :
        if isinstance(ty, BasicType) :
            if ty.const_qualified :
                return self.keyword2Html("const") + " " + self.keyword2Html(ty.ty)
            return self.keyword2Html(ty.ty)

        if isinstance(ty, RValueRefType) :
            assert ty.to
            if ty.const_qualified :
                return self.keyword2Html("const") + " " + self.type2Html(ty.to) + self.keyword2Html("&&")
            return self.type2Html(ty.to) + self.keyword2Html("&&")

        if isinstance(ty, LValueRefType) :
            assert ty.to
            if ty.const_qualified :
                return self.keyword2Html("const") + " " + self.type2Html(ty.to) + self.keyword2Html("&")
            return self.type2Html(ty.to) + self.keyword2Html("&")

        if isinstance(ty, PointerType) :
            assert ty.to
            if ty.const_qualified :
                return self.keyword2Html("const") + " " + self.type2Html(ty.to) + self.keyword2Html("*")
            return self.type2Html(ty.to) + self.keyword2Html("*")

        # TODO conetxt
        if isinstance(ty, ClassType) :
            assert ty.classSymbol
            if ty.const_qualified :
                return self.keyword2Html("const") + " " + self.symbolCompleteHtmlName(ty.classSymbol)
            return self.symbolCompleteHtmlName(ty.classSymbol)

        if isinstance(ty, TypeAliasType) :
            assert ty.aliasSymbol
            if ty.const_qualified :
                return self.keyword2Html("const") + " " + self.symbolCompleteHtmlName(ty.aliasSymbol)
            return self.symbolCompleteHtmlName(ty.aliasSymbol)

        if isinstance(ty, AutoType) :
            assert ty.deduced
            if ty.const_qualified :
                # TODO span-to-change
                return "(" + self.keyword2Html("const") + " " + self.symbolCompleteHtmlName(ty.deduced) + " " + self.tag("span-to-change", "d.f." + self.tag("span", "<b>d</b>educed <b>f</b>rom", 'class="tooltiptext"'), 'class="tooltip"') + " " + self.keyword2Html("const") + " " + self.keyword2Html("auto") + ")"
            return "(" + self.symbolCompleteHtmlName(ty.deduced) + " d.f. " + self.keyword2Html("auto") + ")"
        
        return "type2Html_ERROR"

    def keyword2Html(self, ky : str) -> str :
        href : str | list[str] = self.generate_keyword_href(ky)

        if ky == "&" or ky == "&&" or ky == "*" :
            # TODO better without span
            if href :
                # TODO change span
                return self.code_span_tag("span-to-change", ky, href)
                return self.tag("a", ky, f'href="{href}" class="code-link"')
            else :
                return ky

        return self.code_span_tag("code-keyword", ky, href)

    def tmp_page(self, doc : ParsedDocument) -> str :
        html : str = ""
        for s in doc.symbolTable.symbols :
            html += self.symbolCompleteHtmlName(s) + "<br>"
        html += "<hr>"
        for s in doc.symbolTable.symbols :
            if isinstance(s, MethodSymbol) :
                ret = typing.cast(cxx.FuncArg, s.ret)
                assert ret.typ
                html += self.type2Html(ret.typ) + " ciao<br>"
        html += "<hr>"
        for s in doc.symbolTable.symbols :
            if s.declaration :
                if s.declaration.node.file() == doc.filePath :
                    html += self.symbolCompleteHtmlName(s) + "<br>"
        return f"""
    <!DOCTYPE html>
    <html lang="en">
    <head class="light-mode">
        <meta charset="UTF-8">
        <meta http-equiv="X-UA-Compatible" content="IE=edge">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Document</title>

        <link rel="stylesheet" href="https://lucaciucci99.com/css/style.css">
        <link rel="stylesheet" href="./style.css">

        <script async src="https://cdn.jsdelivr.net/npm/tex-math@latest/dist/tex-math.js"></script>
    </head>
    <body light-mode->

        <lc-content>
            <section>
                <article>
                    <article-page>
                        <pre><code>{html}</code></pre>
                        <hr>
                    </article-page>
                    <article-page>
                    </article-page>
                </article>
            </section>
        </lc-content>
    </body>
    </html>
        """

    def html_article(self, content : str) -> str :
        return self.tag("article", content)

    def html_article_page(self, content : str) -> str :
        return self.tag("article-page", content)

    def get_relative_path(self, symbol : Symbol) -> str :
        if symbol.parent :
            return path.join(self.get_relative_path(symbol.parent), symbol.name)
        return symbol.name

    def write_namespace_page(self, ns : NamespaceSymbol, folder : str) -> None :
        #folder2 = path.join(folder, self.get_relative_path(ns))
        folder2 = folder
        self.m_currentFileFolder = folder2
        filePath = path.join(folder2, "index.html")
        os.makedirs(os.path.dirname(filePath), exist_ok=True)
        with open(filePath, 'wb') as f :
            html : str = ""

            # title
            html += f"<h1>Namespace {self.symbolCompleteHtmlName(ns)}</h1>"

            classes : str = ""
            for symbol in ns.children :
                if isinstance(symbol, ClassSymbol) :
                    classes += self.tag("tr",
                        self.tag("td", self.keyword2Html("class"), 'style="width:15%;text-align:right;"') + self.tag("td", self.symbolHtmlName(symbol), 'style="width:15%"') + self.tag("td", symbol.documentationString.brief)
                        )
                    #classes += f"<li>{} {symbol.documentationString.brief}</li>"
            if classes :
                html += f"<h2>Classes</h2><hr>"
                html += self.tag("table", classes, 'class="basic-table"')
                #html += f"<ul><code>{classes}</code></ul>"

            f.write(self.html_page_base(self.html_article(self.html_article_page(html))).encode('utf-8'))

    def write_class_page(self, cl : ClassSymbol, folder : str) -> None :
        self.m_currentFileFolder = folder
        filePath = path.join(folder, "index.html")
        os.makedirs(os.path.dirname(filePath), exist_ok=True)
        with open(filePath, 'wb') as f :
            html : str = ""

            # title
            html += f"<h1>Class {self.symbolHtmlName(cl)}</h1>"
            
            # brief description
            if cl.documentationString.docStr and cl.documentationString.docStr.rstrip(" \n\t*") != cl.documentationString.brief :
               html += f"""<p>{cl.documentationString.brief} {self.tag("a", "More...", 'href="#detailed-description"')}</p>"""
            else :
                # brief description only
                html += f"<p>{cl.documentationString.brief}</p>"

            # class full name
            html += f"""<div class=\"p\">Full symbol: {self.code_tag(self.keyword2Html('class') + " " + self.symbolCompleteHtmlName(cl))}</p>"""

            # methods
            rows : str = ""
            for method in cl.children :
                if isinstance(method, MethodSymbol) :
                    if method.access_specifier == AccessSpecifier.PUBLIC or True:
                        rows += self.method2tr(method)

            methods_table = self.methods_table(cl.children, [AccessSpecifier.PUBLIC])
            if methods_table :
                html += self.tag("h2", "Public methods")
                html += methods_table

            methods_table = self.methods_table(cl.children, [AccessSpecifier.PROTECTED])
            if methods_table :
                html += self.tag("h2", "Protected methods")
                html += methods_table

            methods_table = self.methods_table(cl.children, [AccessSpecifier.PRIVATE])
            if methods_table :
                html += self.tag("h2", "Private methods")
                html += methods_table

            # detailed description
            if cl.documentationString.docStr and cl.documentationString.docStr.rstrip(" \n\t*") != cl.documentationString.brief :
                html += "<br><br>" + self.tag("h2", "Detailed description", 'id="detailed-description"')
                 # brief description and long description
                doc = DocBlock()
                doc.parse(cl.documentationString.docStr)
                html += doc.brief + "<br>" + doc.long


            f.write(self.html_page_base(html).encode('utf-8'))
            #f.write(self.html_page_base(self.tag("lc-sidebar", "Lorem ipsum dolor sit amet consectetur adipisicing elit. Qui deleniti recusandae itaque cum ipsa obcaecati quod ex tempora quo quibusdam non ipsam suscipit, velit veniam id, corporis, in reprehenderit cumque.") + self.html_article(self.html_article_page(html))).encode('utf-8'))

    def methods_table(self, children : list[Symbol], aps : list[AccessSpecifier]) -> str :
        rows : str = ""
        for method in children :
            if isinstance(method, MethodSymbol) :
                found : bool = len(aps) <= 0
                for ap in aps :
                    if method.access_specifier == ap :
                        found = True

                if found :
                    rows += self.method2tr(method)

        return self.tag("table", rows, 'class="basic-table"') if rows else ""

    def method2tr(self, method : MethodSymbol) -> str :
        row : str = ""

        # return type
        ret = typing.cast(cxx.FuncArg, method.ret)
        assert ret.typ
        row += self.tag("td", self.code_tag(self.type2Html(ret.typ)), 'class="return-type"')

        # function
        func : str = ""
        # TODO method.args cast!
        func += self.code_tag(self.symbolHtmlName(method) + "(" + self.named_arg_list(typing.cast(list[cxx.FuncArg], method.args)) + ")" + (" " + self.keyword2Html("const") if method.const_qualified else ""))

        # description
        doc_html : str = ""
        if method.documentationString.docStr and method.documentationString.docStr.rstrip(" \n\t*") != method.documentationString.brief:
            doc = doc_utils.DocBlock()
            doc.parse(method.documentationString.docStr)
            doc_html += self.tag("details",
                self.tag("summary", doc.brief) +
                doc.long +
                (self.tag("h4", "Params") + self.named_args_doc(typing.cast(list[cxx.FuncArg], method.args), doc.params) if len(method.args) > 0 else "")
                )
            # TODO params
        else :
            doc_html += method.documentationString.brief

        row += self.tag("td", func + "<br>" + doc_html)

        return self.tag("tr", row, f'id="{self.generete_symbol_id(method)}"')

    def named_arg_list(self, args : list[cxx.FuncArg]) -> str : 
        result : str = ""
        for i in range(len(args)) :
            arg = args[i]

            # type
            assert arg.typ
            result += self.type2Html(arg.typ) + " "

            # name
            # TODO link for param
            result += self.tag("code-mvar", arg.name)

            if i < len(args) - 1 :
                result += ", "
        return result

    def named_args_doc(self, args : list[cxx.FuncArg], paramsdoc : list[ParamDocBlock]) -> str :
        if len(args) <= 0 :
            return ""

        rows : str = ""
        for arg in args :
            row : str = ""

            # type
            assert arg.typ
            row += self.tag("td", self.code_tag(self.type2Html(arg.typ)), 'class="type"')

            # find the paramdoc
            paramdoc : ParamDocBlock | None = None
            for pd in paramsdoc :
                if arg.name == pd.name :
                    paramdoc = pd
                    break

            doc_str : str = ""
            
            # name
            doc_str += self.code_tag(self.tag("code-mvar", arg.name))

            # doc
            if paramdoc :
                doc = DocBlock()
                doc.parse(paramdoc.descr)
                doc_str += self.tag("details", self.tag("summary", doc.brief) + doc.long)

            row += self.tag("td", doc_str)
            
            rows += self.tag("tr", row)

        return self.tag("table", rows, 'class="basic-table"')



    def write_symbolTree(self, symbolTree : SymbolsTree, dir : str) -> None :
        dir2 : str = dir
        if symbolTree.symbol :
            dir2 = path.join(dir, symbolTree.symbol.name)

        # write pages
        if isinstance(symbolTree.symbol, NamespaceSymbol) :
            self.write_namespace_page(symbolTree.symbol, dir2)
        if isinstance(symbolTree.symbol, ClassSymbol) :
            self.write_class_page(symbolTree.symbol, dir2)

        for child in symbolTree.children :
            self.write_symbolTree(child, dir2)

    def write_to(self, folder : str) -> None :
        self.m_currentOutFolder = folder
        folder2 = path.join(folder, "symbols")
        self.write_rc_file("style.css", folder)
        self.write_rc_file("code.css", folder)
        self.write_rc_file("index.js", folder)

        self.write_symbolTree(self.symbolTree, folder2)

        return
        for t in self.symbolTree.children :
            s = t.symbol
            if isinstance(s, NamespaceSymbol) :
                self.write_namespace_page(s, folder2)
        pass




"""LC_NOTICE_BEGIN
LC_NOTICE_END"""

#import clang.cindex

class Symbol :
    declaration : Declaration | None = None
    definition : Declaration | None = None

    name : str = ""

    parent : Symbol | None = None
    children : list[Symbol] = []

    def addToParent(self, parent : Symbol) -> None :
        parent.children.append(self)
        self.parent = parent

    def getCompleteName(self) -> str :
        if (self.parent) :
            return self.parent.getCompleteName() + "!!" + self.name
        return self.name



class Declaration :
    ciao : str = "aa"
    node : clang.cindex.Cursor

    symbol : Symbol | None = None

    def __init__(self, node : clang.cindex.Cursor) -> None :
        self.node = node

    def spelling(self) -> str :
        return self.node.spelling

    def __str__(self) -> str:
        return str(self.node.kind) + " " + self.node.spelling

    def __repr__(self) -> str:
        # see https://docs.python.org/3/reference/datamodel.html#object.__repr__
        return self.node.displayname
        return self.node.spelling



from __future__ import annotations
from ast import parse
import pathlib
import time

#from ast import parse
#from sqlite3 import Cursor
#from types import NoneType
from typing import Iterator
#import typing

# https://pypi.org/project/libclang/
# https://libclang.readthedocs.io/en/latest/
# https://sudonull.com/post/907-An-example-of-parsing-C-code-using-libclang-in-Python
#import clang.cindex



#from .lcdoc import *
from lcdoc import *
from cxx import *
from cxx_generator import *
from cxx_target import *
from WebsiteGenerator import WebsiteGenerator

#from src.Declaration import Declaration



#index : clang.cindex.Index = clang.cindex.Index.create()
#tu : clang.cindex.TranslationUnit = index.parse("D:/LUCA/LUCE_STRUTTURATA/develop/StructuredLight/StructuredLight/include/StructuredLight/SLScanner/scans.hpp")




parser = DocumentParser()
#doc = ParsedDocument()
#doc.AST = parser.parse("D:/LUCA/LUCE_STRUTTURATA/develop/StructuredLight/StructuredLight/include/StructuredLight/SLScanner/scans.hpp")
#parser.includes.append(pathlib.Path.absolute(pathlib.Path("./include")))
parser.includes.append("./example/project/include")
parser.inFileOnly = False
doc : ParsedDocument | None = None
doc = parser.parse("./example/project/src/test.cpp")
try :
    #doc = parser.parse("./out/test.cpp")
    pass
except ValueError as ve:
    print("Value Error: ", ve)
except clang.cindex.LibclangError as e:
    print("error: ", e)
except clang.cindex.CompilationDatabaseError as e2:
    print(e2)
except :
    print("error")

#doc.findDeclarations()
#print(doc.declarations)
print(doc.symbolTable.symbols)

generator = Generator()
#print(generator.html(doc))
#with open('./out/index.html', 'wb') as f:
#    f.write(generator.html(doc).encode('utf-8'))

gen = HtmlGenerator()
#for s in doc.symbolTable.symbols :
#    print(gen.symbolHtmlName(s))

gen.add_doc(doc)
#with open('./out/index2.html', 'wb') as f:
    #f.write(gen.tmp_page(doc).encode('utf-8'))
#    pass
gen.write_to("./example/out")

while True :
    gen.doc_dir("./doc", "./doc_out")
    time.sleep(10)

#show_list_CursorKind()